#!/usr/bin/env python
# Copyright 2013 Brett Slatkin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants used by all sub-modules of dpxdt.

* Do not import anything into this file!
* Do not put configuration in this file!
* Do not put secret keys in this file!
"""


CAPTURE_QUEUE_NAME = 'capture'

PDIFF_QUEUE_NAME = 'run-pdiff'

SITE_DIFF_QUEUE_NAME = 'site-diff'
